'''
The PipedViewer class is used to create, send commands, and shutdown viewers 
in this module.  Currently, the only known viewer types are "PipedViewerPQ", 
and "PipedImagerPQ".

This package was developed by the Thermal Modeling and Analysis Project 
(TMAP) of the National Oceanographic and Atmospheric Administration's (NOAA) 
Pacific Marine Environmental Lab (PMEL).
'''

from __future__ import print_function

import sys

import multiprocessing
import threading

WINDOW_CLOSED_MESSAGE = 'Window was closed'

class ErrorMonitor(threading.Thread):
    '''
    Thread to monitor unexpected responses (error messages) sent by 
    a PipedViewer process.  When a response from the PipedViewer is 
    detected, a lock paired with this response pipe is acquired, the 
    response (if still present) is read, and the lock is released. 
    Any error message read is appropriately dealt with and monitoring 
    of the response pipe resumes.
    '''

    def __init__(self, responsepipe, readlock, vprocess):
        '''
        Monitor unexpected responses (error message) from the given 
        response pipe for the PipedViewer.  Prior to reading a response 
        message, the given lock is acquired.  After reading the message, 
        if any was still present, the lock is released, and any error 
        message is handled appropriately.  Exits when the PipedViewer
        process is no longer alive.

        Arguments:
            responsepipe: (multiprocessing.Pipe) provides responses sent by the PipedViewer
            readlock: (threading.RLock) lock associated with the above response pipe
            vprocess: (multiprocessing.Process) PipedViewer process
        '''
        super(ErrorMonitor, self).__init__(group=None, target=None, name='ErrorMonitor')
        self.__responsepipe = responsepipe
        self.__readlock = readlock
        self.__vprocess = vprocess

    def run(self):
        '''
        Monitor unexpected responses (error message) from the PipedViewer 
        response pipe and read lock associated with this instance.  Prior 
        to reading a response message, the lock is acquired.  After reading 
        the message, if any was still present, the lock is released, and 
        any error message is handled appropriately.  Exits when the reponse 
        pipe is closed (when multiprocessing.Pipe.poll raises an Exception).
        '''
        while True:
            try:
                # wait for something to appear or the PipedViewer to exit
                while not self.__responsepipe.poll(0.1):
                    if not self.__vprocess.is_alive():
                        sys.exit(0)
            except Exception:
                # response pipe was closed - quit
                break
            # read the message, if still there, only after acquiring the read lock
            fullerrmsg = ''
            self.__readlock.acquire()
            try:
                try:
                    while self.__responsepipe.poll():
                        errmsg = self.__responsepipe.recv()
                        if fullerrmsg:
                            fullerrmsg += '\n'
                        fullerrmsg += str(errmsg)
                except Exception:
                    # response pipe was closed - deal with anything already read;
                    # next loop through will break and quit
                    pass
            finally:
                self.__readlock.release()
            # deal with the error message, if any
            if fullerrmsg == WINDOW_CLOSED_MESSAGE:
                # TODO: tell the ferret engine that the window was closed 
                #       and maybe don't print anything
                print('\n', file=sys.stderr)
                print(fullerrmsg, file=sys.stderr)
            elif fullerrmsg:
                print('\n', file=sys.stderr)
                print(fullerrmsg, file=sys.stderr)
        sys.exit(0)


class PipedViewer(object):
    '''
    Creates and starts a PipedViewer of one of the supported viewer 
    types.  Provides methods for interacting with the PipedViewer.
    '''
    def __init__(self, viewertype):
        '''
        Create and start a PipedViewer of one of the supported viewer 
        types.  The viewer will probably not be displayed until the 
        {"action":"show"} command is submitted to the viewer.

        Currently supported viewer types are:
            "PipedViewerPQ": PipedViewerPQ using PySide2, PyQt5, or PyQt4
            "PipedImagerPQ": PipedImagerPQ using PySide2, PyQt5, or PyQt4
        '''
        super(PipedViewer, self).__init__()
        (self.__cmndrecvpipe, self.__cmndsendpipe) = multiprocessing.Pipe(False)
        (self.__rspdrecvpipe, self.__rspdsendpipe) = multiprocessing.Pipe(False)
        if viewertype == "PipedViewerPQ":
            try:
                from pipedviewer.pipedviewerpq import PipedViewerPQProcess
            except ImportError:
                raise TypeError("The PQ viewers requires PySide2, PyQt5, or PyQt4")
            self.__vprocess = PipedViewerPQProcess(self.__cmndrecvpipe,
                                                   self.__rspdsendpipe)
        elif viewertype == "PipedImagerPQ":
            try:
                from pipedviewer.pipedimagerpq import PipedImagerPQProcess
            except ImportError:
                raise TypeError("The PQ viewers requires PySide2, PyQt5, or PyQt4")
            self.__vprocess = PipedImagerPQProcess(self.__cmndrecvpipe,
                                                   self.__rspdsendpipe)
        else:
            raise TypeError("Unknown viewer type %s" % str(viewertype))
        self.__vprocess.start()
        self.__rspdrecvlock = threading.RLock()
        self.__errmonitor = ErrorMonitor(self.__rspdrecvpipe, self.__rspdrecvlock, self.__vprocess)
        self.__errmonitor.start()

    def blockErrMonitor(self):
        '''
        Block monitoring of error messages sent from the PipedViewer. 

        Call this method prior to submitting a command which is expected 
        to respond with desired information so this response will not be 
        mistaken as an error message.  After reading the response, call 
        the "resumeErrMonitor" method. 
        '''
        self.__rspdrecvlock.acquire()

    def resumeErrMonitor(self):
        '''
        Resume monitoring of error messages sent from the PipedViewer. 

        Only call this method after calling "blockErrMonitor", which 
        presumably was called prior to submitting a command expected 
        to respond with desired information and reading the response.
        '''
        self.__rspdrecvlock.release()

    def submitCommand(self, cmnd):
        '''
        Submit the command cmnd to the command pipe for the viewer.

        If a response is expected from the command, the "blockErrMonitor" 
        method should be called prior to submitting the command, and 
        the "resumeErrMonitor" method should be called after reading 
        the response.  This prevent the response from being mistaken 
        as an error message.
        '''
        self.__cmndsendpipe.send(cmnd)

    def checkForResponse(self, timeout = 0.0):
        '''
        Check for a response from the viewer.  

        Arguments: 
            timeout: (number) maximum time in seconds to wait for a 
                     response to appear.  Zero (default) does not wait;
                     None waits indefinitely.
        Returns: 
            the response from the viewer, or None if there was no 
            response in the allotted time.
        Raises: 
            IOError: if the response pipe is closed while waiting 
                     for a response to appear.
        '''
        if self.__rspdrecvpipe.poll(timeout):
            myresponse = self.__rspdrecvpipe.recv()
        else:
            myresponse = None
        return myresponse

    def shutdownViewer(self):
        '''
        If the PipedViewer is still alive, submits the command {"action":"exit"}
        and reads any responses.  Waits for the viewer and the error montior to exit.  

        Returns: 
            any responses from the viewer, or 
            an empty string if there was no response (viewer was not alive).
        '''
        closingremarks = ''
        if self.__vprocess.is_alive():
            # block the error monitor
            self.__rspdrecvlock.acquire()
            self.__cmndsendpipe.send({'action':'exit'})
            self.__cmndsendpipe.close()
            try:
                # Read everything from the response pipe until it is closed on the viewer side
                if closingremarks:
                    closingremarks += '\n'
                closingremarks += self.__rspdrecvpipe.recv()
            except Exception:
                pass
            self.__rspdrecvpipe.close()
            self.__rspdrecvlock.release()
        self.__vprocess.join()
        self.__errmonitor.join()
        return closingremarks

    def getViewerExitCode(self):
        '''
        Returns: 
            the viewer process exit code.
        '''
        return self.__vprocess.exitcode


def _testviewers():
    # vertices of a pentagon (roughly) centered in a 1000 x 1000 square
    pentagonpts = ( (504.5, 100.0), (100.0, 393.9),
                    (254.5, 869.4), (754.5, 869.4),
                    (909.0, 393.9),  )
    # create the list of commands to submit
    drawcmnds = []
    drawcmnds.append( { "action":"setTitle", "title":"Tester" } )
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"clear", "color":"black"} )
    drawcmnds.append( { "action":"screenInfo"} )
    drawcmnds.append( { "action":"antialias", "antialias":True } )
    drawcmnds.append( { "action":"resize",
                        "width":500,
                        "height":500 } )
    drawcmnds.append( { "action":"beginView",
                        "viewfracs":{"left":0.0, "right":0.5,
                                     "top":0.5, "bottom":1.0},
                        "clip":True } )
    drawcmnds.append( { "action":"drawRectangle",
                        "left": 5, "right":245, 
                        "top":245, "bottom":495,
                        "fill":{"color":"green", "alpha":128} } )
    mypentapts = [ (.25 * ptx, .25 * pty + 250) for (ptx, pty) in pentagonpts ]
    drawcmnds.append( { "action":"drawPolygon",
                        "points":mypentapts,
                        "fill":{"color":"blue"},
                        "outline":{"color":"black",
                                   "width": 5,
                                   "style":"solid",
                                   "capstyle":"round",
                                   "joinstyle":"round" } } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=480",
                        "font":{"family":"Times", "size":50},
                        "fill":{"color":"red"},
                        "location":(50,480) } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=430",
                        "font":{"family":"Times", "size":50},
                        "fill":{"color":"red"},
                        "location":(50,430) } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=380",
                        "font":{"family":"Times", "size":50},
                        "fill":{"color":"red"},
                        "location":(50,380) } )
    drawcmnds.append( { "action":"drawText",
                        "text":"y=330",
                        "font":{"family":"Times", "size":50},
                        "fill":{"color":"red"},
                        "location":(50,330) } )
    drawcmnds.append( { "action":"endView" } )
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"createSymbol",
                        "name": "bararrow",
                        "pts": ( (-50,50), (-10,10),
                                 (-999, -999),
                                 (50,0), (50,50), (0,50),
                                 (-999, -999),
                                 (0,-10), (20,-30), (10,-30), (10,-50), (-10,-50), (-10,-30), (-20,-30), (0,-10), ),
                        "fill": False } )
    drawcmnds.append( { "action":"beginView",
                        "viewfracs":{"left":0.0, "right":1.0,
                                     "top":0.0, "bottom":1.0},
                        "clip":True } )
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (100,  50),
                                   (100, 150),
                                   (100, 250),
                                   (100, 350),
                                   (100, 450) ),
                        "symbol":".",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (150,  50),
                                   (150, 150),
                                   (150, 250),
                                   (150, 350),
                                   (150, 450) ),
                        "symbol":"o",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (200,  50),
                                   (200, 150),
                                   (200, 250),
                                   (200, 350),
                                   (200, 450) ),
                        "symbol":"+",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (250,  50),
                                   (250, 150),
                                   (250, 250),
                                   (250, 350),
                                   (250, 450) ),
                        "symbol":"x",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (300,  50),
                                   (300, 150),
                                   (300, 250),
                                   (300, 350),
                                   (300, 450) ),
                        "symbol":"*",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (350,  50),
                                   (350, 150),
                                   (350, 250),
                                   (350, 350),
                                   (350, 450) ),
                        "symbol":"bararrow",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (400,  50),
                                   (400, 150),
                                   (400, 250),
                                   (400, 350),
                                   (400, 450) ),
                        "symbol":"^",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawPoints",
                        "points":( (450,  50),
                                   (450, 150),
                                   (450, 250),
                                   (450, 350),
                                   (450, 450) ),
                        "symbol":"#",
                        "size":20,
                        "color":"magenta" })
    drawcmnds.append( { "action":"drawMultiline",
                        "points":( (350,  50),
                                   (200, 150),
                                   (400, 250),
                                   (300, 350),
                                   (150, 250),
                                   (100, 450) ),
                        "pen": {"color":"white",
                                "width":3,
                                "style":"dash",
                                "capstyle":"round",
                                "joinstyle":"round"} } )
    drawcmnds.append( { "action":"endView" } )
    drawcmnds.append( { "action":"show" } )
    testannotations = ( "The 1<sup>st</sup> CO<sub>2</sub> annotations line",
                        "Another line with <i>lengthy</i> details that go on and on " + \
                        "and on and should wrap to a 2<sup>nd</sup> annotation line",
                        "<b>Final</b> annotation line" )

    # Test each known viewer.
    for viewername in ( "PipedViewerPQ", ):
        print("Testing Viewer %s" % viewername)
        # create the viewer
        pviewer = PipedViewer(viewername)
        mydrawcmnds = drawcmnds[:]
        mydrawcmnds.append( { "action":"save",
                              "filename":viewername + "_test.pdf",
                              "vectsize":{"width":7.0, "height":7.0},
                              "rastsize":{"width":750, "height":750},
                              "annotations":testannotations } )
        mydrawcmnds.append( { "action":"save",
                              "filename":viewername + "_test.png",
                              "vectsize":{"width":7.0, "height":7.0},
                              "rastsize":{"width":750, "height":750},
                              "annotations":testannotations } )

        # submit the commands, pausing after each "show" command
        for cmd in mydrawcmnds:
            print("Command: %s" % str(cmd))
            pviewer.submitCommand(cmd)
            if cmd["action"] == "show":
                if sys.version_info[0] > 2:
                    input("Press Enter to continue")
                else:
                    raw_input("Press Enter to continue")
        # end of the commands - shut down and check return value
        response = pviewer.shutdownViewer()
        if response:
            print("Closing remarks: %s" % str(response))
        result = pviewer.getViewerExitCode()
        if result != 0:
            print("Exit code %i from %s" % (result, viewername))
        else:
            print("Done with %s" % viewername)

if __name__ == "__main__":
    _testviewers()



'''
PyFerret external function providing data partitioned into summed 
pieces along the ensemble axis.  The data at ensemble value 'm' 
is the data explained by the 'm' most-significant Empirical Orthogonal 
Functions (EOFs) and their corresponding Time Amplitude Functions 
(TAFs).

@author: Karl Smith
'''

from __future__ import print_function

import numpy
import pyferret
import pyferret.eofanal as eofanal

def ferret_init(efid):
    '''
    Initializes the eofdatasum function. 
    '''
    init_dict = { }
    init_dict["numargs"] = 3
    init_dict["descript"] = "Partitions data into EOF * TAF sums " + \
        "(total data explained) along the ensemble axis"
    init_dict["argnames"] = ("Data",
                             "MinSignif",
                             "MaxEOFs")
    init_dict["argdescripts"] = (
        "Time-location data; defined on T and one or more of X, Y, Z",
        "Minimum fraction-of-data-explained considered significant",
        "Maximum number of EOFs to consider (-1 for all signficant EOFS)")
    init_dict["argtypes"] = (pyferret.FLOAT_ARRAY,
                             pyferret.FLOAT_ONEVAL,
                             pyferret.FLOAT_ONEVAL)
    # X, Y, Z, and T match input; E axis added as an abstract axis
    axes = [ pyferret.AXIS_IMPLIED_BY_ARGS ] * pyferret.MAX_FERRET_NDIM
    axes[pyferret.E_AXIS] = pyferret.AXIS_ABSTRACT
    axes[pyferret.F_AXIS] = pyferret.AXIS_DOES_NOT_EXIST;
    init_dict["axes"] = axes
    part_influence = [ True ] * pyferret.MAX_FERRET_NDIM
    part_influence[pyferret.E_AXIS] = False
    part_influence[pyferret.F_AXIS] = False
    no_influence = [ False ] * pyferret.MAX_FERRET_NDIM
    init_dict["influences"] = (part_influence,
                               no_influence,
                               no_influence)
    init_dict["piecemeal"] =  [ False ] * pyferret.MAX_FERRET_NDIM

    return init_dict


def ferret_result_limits(efid):
    '''
    Provides the bounds of the E abstract axis.  
    The maximum number of EOFs is the number of locations. 
    '''
    maxeofs = pyferret.get_arg_one_val(efid, pyferret.ARG3)
    if maxeofs > 0.95:
        maxeofs = int(maxeofs + 0.1)
    else:
        maxeofs = 1
        for axis in (pyferret.X_AXIS, pyferret.Y_AXIS, pyferret.Z_AXIS):
            axis_info = pyferret.get_axis_info(efid, pyferret.ARG1, axis);
            if axis_info:
                npts = axis_info.get("size", -1)
                if npts > 0:
                    maxeofs *= npts
    result_limits = [ None ] * pyferret.MAX_FERRET_NDIM
    result_limits[pyferret.E_AXIS] = (1, maxeofs+1)
    return result_limits


def ferret_compute(efid, result, result_bdf, inputs, input_bdfs):
    '''
    Assign result with the partial sums of EOF * TAF (data explained
    by the most significant EOFs) up to the number of significant 
    EOFs.  The X,Y,Z,T data is given in inputs[0], the minimum 
    fraction-of-data-explained considered significant is given as a 
    single value in inputs[1].
    '''
    # verify no ensemble or forecast axis on the input data
    if inputs[pyferret.ARG1].shape[pyferret.E_AXIS] > 1:
        raise ValueError("Input data cannot have an ensemble axis")
    if inputs[pyferret.ARG1].shape[pyferret.F_AXIS] > 1:
        raise ValueError("Input data cannot have a forecast axis")
    # number of time steps in the input data
    ntime = inputs[pyferret.ARG1].shape[pyferret.T_AXIS]
    if ntime < 2:
        raise ValueError("Input data time axis does not exist or is a singleton")
    # Get the minimum fraction-data-explained and verify it is reasonable
    min_signif = float(inputs[1])
    if (min_signif < 1.0E-6) or (min_signif > (1.0 - 1.0E-6)):
        raise ValueError("MinSignif must be in [0.000001, 0.999999]")
    # Get the maximum number of EOFs to consider
    maxeofs = float(inputs[2])
    if maxeofs > 0.95:
        maxeofs = int(maxeofs + 0.1)
    else:
        maxeofs = -1
    # Get the mask of where the data is defined
    defined_data = ( numpy.fabs(inputs[pyferret.ARG1] - 
                                input_bdfs[pyferret.ARG1]) >= 1.0E-5 )
    # Get the mask of where the data is defined for every time step
    defd_mask = numpy.logical_and.reduce(defined_data, axis=pyferret.T_AXIS)
    for t in range(ntime):
        defined_data[:, :, :, t] = defd_mask 
    # Convert to time-location (a 2-D array), 
    # eliminating locations with missing time steps, 
    # and taking the transpose so time is the first axis
    timeloc = inputs[pyferret.ARG1][defined_data]\
                    .reshape((-1, ntime)).T
    # Create the EOFAnalysis object and analyze the data
    eofs = eofanal.EOFAnalysis(timeloc)
    eofs.setminsignif(min_signif)
    eofs.analyze()
    # Initialize the result to all-undefined
    result[:] = result_bdf
    # Remove the E and F singleton axes from the defined mask
    defined_data = defined_data.reshape(defined_data.shape[:4])
    # Assign the sums of the EOF-TAF products for the significant EOFs
    # The values at m=0 are the time-series averages
    numeofs = eofs.numeofs()
    if (maxeofs > 0) and (numeofs > maxeofs):
        numeofs = maxeofs
    for k in range(numeofs+1):
        timeloc_sum = eofs.dataexplained(k)
        loctime_sum = numpy.array(timeloc_sum.T).reshape(-1)
        result[:, :, :, :, k, 0][defined_data] = loctime_sum
    # The EOF-TAF sums for insignificant EOFs are left as undefined 
    return


if __name__ == "__main__":
    # Just verify ferret_init does not raise an error
    defdict = ferret_init(0)
    # Create some data
    yaxis = numpy.linspace(-80.0, 80.0, 17)
    zaxis = numpy.linspace(0.0, 250.0, 6)
    taxis = numpy.linspace(0.0, 8760.0, 25)
    ydata = 20.0 * numpy.square(numpy.cos(numpy.deg2rad(yaxis)))
    zdata = 1.0 / numpy.log10(5.0 * zaxis + 10.0)
    yzdata = numpy.outer(ydata, zdata)
    tdata = (taxis / 87600.0) + 1.0
    yztdata = numpy.outer(yzdata, tdata).reshape((1, 17, 6, 25, 1, 1))
    ydata = numpy.square(numpy.sin(2.0 * numpy.deg2rad(yaxis)))
    yzdata = numpy.outer(ydata, zdata)
    tdata = -5.0 * numpy.cos((tdata - 2190.0) * numpy.pi / 4380.0)
    yztdata += numpy.outer(yzdata, tdata).reshape((1, 17, 6, 25, 1, 1))
    yztdata += 6.0
    print("time series at Y = 0.0, Z = 0.0")
    print(str(yztdata[0, 8, 0, :, 0, 0]))
    print("depth series at Y = 0.0, T = start of April")
    print(str(yztdata[0, 8, :, 6, 0, 0]))
    print("latitude series at Z = 0.0, T = start of April")
    print(str(yztdata[0, :, 0, 6, 0, 0]))
    # Create the result array and the other ferret_compute arguments
    result = numpy.zeros((1, 17, 6, 25, 17*6, 1))
    resbdf = numpy.array([1.0E20])
    inputs = (yztdata, 0.001, -1)
    inpbdfs = (-1.0E34, -1.0E34, -1.0E34)
    # Check ferret_compute
    ferret_compute(0, result, resbdf, inputs, inpbdfs)
    lastone = 0
    for m in range(17 * 6):
        if numpy.allclose(result[:, :, :, :, m, 0], resbdf):
            break;
        print("EOF-TAF sum %d" % m)
        print("    time series at Y = 0.0, Z = 0.0")
        print("    " + str(result[0, 8, 0, :, m, 0]))
        print("    depth series at Y = 0.0, T = start of April")
        print("    " + str(result[0, 8, :, 6, m, 0]))
        print("    latitude series at Z = 0.0, T = start of April")
        print("    " + str(result[0, :, 0, 6, m, 0]))
        lastone = m
    if numpy.allclose(result[:, :, :, :, lastone, 0], 
                     yztdata[:, :, :, :, 0, 0]):
        print("sum %d all close to input data" % lastone)
    else:
        print("sum %d different from input data" % lastone)




'''
Empirical-orthogonal-function analysis functions designed for use in PyFerret, 
especially for Python external functions for PyFerret.

@author: Karl Smith
'''

# Import classes given in modules in this package so they are all seen here.
from eofanalysis import InvalidStateError, EOFAnalysis




# To be run in python after importing and starting pyferret
# such as from running "pyferret -python"

from __future__ import print_function

import sys ; sys.ps1 = '' ; sys.ps2 = ''
print()

print(">>> coads = pyferret.FerDSet('coads_climatology.cdf')")
coads = pyferret.FerDSet('coads_climatology.cdf')
print(">>> str(coads.sst)")
print(str(coads.sst))
print(">>> coads.sst.showgrid()")
print(coads.sst.showgrid())
print(">>> del coads.sst")
del coads.sst

print(">>> coads_uw  = pyferret.FerDSet('coads_uw.nc')")
coads_uw  = pyferret.FerDSet('coads_uw.nc')
print(">>> coads_uw.uwnd.showgrid()")
coads_uw.uwnd.showgrid()

print(">>> coads_uw.mywnd = (coads_uw.uwnd**2 + coads_uw.vwnd**2)**0.5")
coads_uw.mywnd = (coads_uw.uwnd**2 + coads_uw.vwnd**2)**0.5
print(">>> dir(coads_uw)")
dir(coads_uw)
print(">>> print repr(coads_uw.mywnd)")
print(repr(coads_uw.mywnd))
print(">>> coads_uw.mywnd.showgrid()")
coads_uw.mywnd.showgrid()

print(">>> print repr(coads_uw.mywnd.data)")
print(repr(coads_uw.mywnd.data))
print(">>> print repr(coads_uw.mywnd.grid)")
print(repr(coads_uw.mywnd.grid))
print(">>> coads_uw.mywnd.load()")
coads_uw.mywnd.load()
print(">>> coads_uw.mywnd.data.shape")
coads_uw.mywnd.data.shape
print(">>> print repr(coads_uw.mywnd.data[2:5,2:5,0,0,0,0])")
print(repr(coads_uw.mywnd.data[2:5,2:5,0,0,0,0]))
print(">>> print repr(coads_uw.mywnd.grid)")
print(repr(coads_uw.mywnd.grid))
print(">>> coads_uw.mywnd.unload()")
coads_uw.mywnd.unload()

print(">>> coads.sstregrid = coads.sst.regrid(coads_uw.mywnd, pyferret.REGRID_AVERAGE)")
coads.sstregrid = coads.sst.regrid(coads_uw.mywnd, pyferret.REGRID_AVERAGE)
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> coads.sstregrid.showgrid()")
coads.sstregrid.showgrid()
print(">>> coads.sstregrid.load()")
coads.sstregrid.load()
print(">>> coads.sstregrid.data.shape")
coads.sstregrid.data.shape
print(">>> print repr(coads.sstregrid.data[2:5,2:5,0,0,0,0])")
print(repr(coads.sstregrid.data[2:5,2:5,0,0,0,0]))
print("pyferret.run('LIST /X=55W:51W /Y=5N:9N /L=1 SST[d=coads_climatology]')")
pyferret.run('LIST /X=55W:51W /Y=5N:9N /L=1 SST[d=coads_climatology]')

print(">>> coads_uw.show()")
coads_uw.show()
print(">>> del coads_uw.mywnd")
del coads_uw.mywnd
print(">>> coads_uw.show()")
coads_uw.show()

print(">>> pyferret.anondset.sstcopy = coads.sst")
pyferret.anondset.sstcopy = coads.sst
print(">>> pyferret.anondset.show()")
pyferret.anondset.show()
print(">>> print repr(pyferret.anondset.sstcopy)")
print(repr(pyferret.anondset.sstcopy))
print(">>> print repr(pyferret.anondset.sstcopy._isfilevar)")
print(repr(pyferret.anondset.sstcopy._isfilevar))
print(">>> print repr(pyferret.anondset.sstcopy._requires)")
print(repr(pyferret.anondset.sstcopy._requires))
print(">>> print repr(coads.sst)")
print(repr(coads.sst))
print(">>> print repr(coads.sst._isfilevar)")
print(repr(coads.sst._isfilevar))
print(">>> print repr(coads.sst._requires)")
print(repr(coads.sst._requires))
print(">>> pyferret.anondset.close()")
pyferret.anondset.close()
print(">>> pyferret.anondset.show()")
pyferret.anondset.show()

print(">>> coads.sst2 = coads.sst['15-FEB']")
coads.sst2 = coads.sst['15-FEB']
print(">>> coads.sst2.showgrid()")
coads.sst2.showgrid()
print(">>> coads.sst2.load()")
coads.sst2.load()
print(">>> coads.sst2.data.shape")
coads.sst2.data.shape
print(">>> print repr(coads.sst2.data[2:5,23:26,0,0,0,0])")
print(repr(coads.sst2.data[2:5,23:26,0,0,0,0]))
print("pyferret.run('LIST /X=25E:29E /Y=43S:39S /L=2 SST[d=coads_climatology]')")
pyferret.run('LIST /X=25E:29E /Y=43S:39S /L=2 SST[d=coads_climatology]')

print(">>> coads.sst2 = coads.sst['43S':'39S','25E':'29E',:,1]")
coads.sst2 = coads.sst['43S':'39S','25E':'29E',:,1]
print(">>> print repr(coads.sst2)")
print(repr(coads.sst2))
print(">>> coads.show(brief=False)")
coads.show(brief=False)
print(">>> coads.sst2.load()")
coads.sst2.load()
print(">>> coads.sst2.data.shape")
coads.sst2.data.shape
print(">>> print repr(coads.sst2.data[:,:,0,0,0,0])")
print(repr(coads.sst2.data[:,:,0,0,0,0]))


print(">>> pyferret.showdata(brief=False)")
pyferret.showdata(brief=False)
print(">>> coads_uw.close()")
coads_uw.close()
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> coads.close()")
coads.close()
print(">>> pyferret.showdata()")
pyferret.showdata()




'''
Represents Ferret variables in Python.
'''

import numbers
import pyferret

# common regridding methods
REGRID_LINEAR = "@LIN"
REGRID_AVERAGE = "@AVE"
REGRID_ASSOCIATE = "@ASN"
REGRID_MEAN = "@BIN"
REGRID_NEAREST = "@NRST"
REGRID_MIN = "@MIN"
REGRID_MAX = "@MAX"
REGRID_EXACT = "@XACT"

_ADDED_ATTRIBUTES = ('data', 'grid', 'missval', 'unit')

class FerVar(object):
    '''
    Ferret variable object
    '''

    def __init__(self, defn=None, title=None):
        '''
        Creates a Ferret variable with (optionally) a title and a Ferret-syntax definition 
            defn (string): Ferret-syntax definition of the variable
            title (string): title (descriptive long name) for this variable
        '''
        # Record or generate the definition, or set to an empty string
        if defn:
            if not isinstance(defn, str):
                raise ValueError("defn is not a string")
            self._definition = defn
        else:
            self._definition = ''
        # Name of the variable in the dataset
        self._varname = ''
        # Name of the dataset
        self._dsetname = ''
        # Record the title for this variable, or am empty string if not given
        self.settitle(title)
        # Is this a file variable?
        self._isfilevar = False
        # The list of uppercase _varname's that are know to be used
        # in the definition.  This list is not guaranteed to be complete 
        # and is not used in comparisons.
        self._requires = set()
        # Call the unload method to create and set the defaults for 
        # _datagrid, _dataarray, _dataunit, and _missingvalue.
        #     _datagrid is a FerGrid describing the Ferret grid for the variable.
        #     _dataarray is a NumPy ndarray contains the Ferret data for the variable.
        #     _dataunit is a string given the unit of the data
        #     _missingvalue is the missing value used for the data
        self.unload()


    def copy(self):
        '''
        Return an anonymous copy (only the definition is copied) of this FerVar.
        '''
        newvar = FerVar(defn=self._definition)
        newvar._requires.update(self._requires)
        return newvar


    def settitle(self, title):
        '''
        Assigns the title (long descriptive name) for this FerVar.  If this
        variable is defined in Ferret, the title for the Ferret variable is
        also updated.
            title (string): title to assign
        Raises ValueError if title is not a string or if there is a problem 
            updating the title in Ferret
        '''
        if title:
            if not isinstance(title, str):
                raise ValueError("title is not a string")
            self._title = title
        else:
            self._title = ''
        if self._varname:
            cmdstr = 'SET VAR/TITLE="%s" %s' % (self._title, self.fername())
            (errval, errmsg) = pyferret.run(cmdstr)
            if errval != pyferret.FERR_OK:
                raise ValueError('problems updating the variable title in Ferret for ' + \
                                 '%s to "%s": %s' % (self.fername(), self._title, errmsg))
 

    def fername(self):
        ''' 
        Returns the Ferret name for this variable; namely,
            <_varname>[d=<_dsetname>]
        if _dsetname is given; otherwise, just
            <_varname>
        Raises ValueError if _varname is not defined
        '''
        if not self._varname:
            raise ValueError('this FerVar does not contain a Ferret variable name')
        if self._dsetname:
            fername = '%s[d=%s]' % (self._varname, self._dsetname)
        else:
            fername = '%s' % self._varname
        return fername


    def __repr__(self):
        '''
        Representation of this FerVar
        '''
        infostr = "FerVar(varname='%s', dsetname='%s', title = '%s', defn='%s')" \
                  % (self._varname, self._dsetname, self._title, self._definition)
        return infostr


    def __del__(self):
        '''
        Removes this variable, if possible, from Ferret.
        Any error are ignored.
        '''
        # Try to remove from Ferret but ignore errors
        try:
            self._removefromferret()
        except Exception:
            pass


    def __cmp__(self, other):
        '''
        FerVars are ordered alphabetically, case-insensitive, first by 
        the Ferret variable name, then by the dataset name, title, and 
        finally by the definition.  (Used by the "rich comparison" methods.)
        '''
        if not isinstance(other, FerVar):
            raise NotImplementedError('other is not a FerVar')
        supper = self._varname.upper()
        oupper = other._varname.upper()
        if supper < oupper:
            return -1
        if supper > oupper:
            return 1
        supper = self._dsetname.upper()
        oupper = other._dsetname.upper()
        if supper < oupper:
            return -1
        if supper > oupper:
            return 1
        supper = self._title.upper()
        oupper = other._title.upper()
        if supper < oupper:
            return -1
        if supper > oupper:
            return 1
        supper = self._definition.upper()
        oupper = other._definition.upper()
        if supper < oupper:
            return -1
        if supper > oupper:
            return 1
        return 0


    def __eq__(self, other):
        '''
        Two FerVars are equal if all of the following are True:
            they have the same Ferret variable name,
            they have the same dataset name, 
            they have the same title, and
            they have the same definition.
        All these comparisons are case-insensitive.
        '''
        try:
            return ( self.__cmp__(other) == 0 )
        except NotImplementedError:
            return NotImplemented


    def __ne__(self, other):
        '''
        Two FerVars are not equal if any of the following are True:
            they have different Ferret variable names,
            they have different dataset names,
            they have different titles, or
            they have different definitions.
        All these comparisons are case-insensitive.
        '''
        try:
            return ( self.__cmp__(other) != 0 )
        except NotImplementedError:
            return NotImplemented


    def __lt__(self, other):
        '''
        FerVars are ordered alphabetically, case-insensitive, first by 
        the Ferret variable name, then by the dataset name, title, and 
        finally by the definition.
        '''
        try:
            return ( self.__cmp__(other) < 0 )
        except NotImplementedError:
            return NotImplemented


    def __le__(self, other):
        '''
        FerVars are ordered alphabetically, case-insensitive, first by 
        the Ferret variable name, then by the dataset name, title, and 
        finally by the definition.
        '''
        try:
            return ( self.__cmp__(other) <= 0 )
        except NotImplementedError:
            return NotImplemented


    def __gt__(self, other):
        '''
        FerVars are ordered alphabetically, case-insensitive, first by 
        the Ferret variable name, then by the dataset name, title, and 
        finally by the definition.
        '''
        try:
            return ( self.__cmp__(other) > 0 )
        except NotImplementedError:
            return NotImplemented


    def __ge__(self, other):
        '''
        FerVars are ordered alphabetically, case-insensitive, first by 
        the Ferret variable name, then by the dataset name, title, and 
        finally by the definition.
        '''
        try:
            return ( self.__cmp__(other) >= 0 )
        except NotImplementedError:
            return NotImplemented


    def __nonzero__(self):
        '''
        Returns False if the Ferret variable name, dataset name, title, 
        and definition are all empty.  (For Python2.x)
        '''
        if self._varname:
            return True
        if self._dsetname:
            return True
        if self._title:
            return True
        if self._definition:
            return True
        return False


    def __bool__(self):
        '''
        Returns False if the Ferret variable name, dataset name, title 
        and definition are all empty.  (For Python3.x)
        '''
        return self.__nonzero__()


    def __add__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the sum of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the sum of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) + (%s)' % (self._definition, other._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '(%s) + %s' % (self._definition, str(other))
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __radd__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the sum of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the sum of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) + (%s)' % (other._definition, self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '%s + (%s)' % (str(other), self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __sub__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the difference (self - other) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the difference (self - other) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) - (%s)' % (self._definition, other._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '(%s) - %s' % (self._definition, str(other))
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __rsub__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the difference (other - self) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the difference (other - self) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) - (%s)' % (other._definition, self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '%s - (%s)' % (str(other), self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __mul__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the product of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the product of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) * (%s)' % (self._definition, other._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '(%s) * %s' % (self._definition, str(other))
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __rmul__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the product of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the product of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) * (%s)' % (other._definition, self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '%s * (%s)' % (str(other), self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __truediv__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the quotient (self / other) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the quotient (self / other) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        (For Python3.x)
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) / (%s)' % (self._definition, other._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '(%s) / %s' % (self._definition, str(other))
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __rtruediv__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the quotient (other / self) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the quotient (other / self) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        (For Python3.x)
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) / (%s)' % (other._definition, self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '%s / (%s)' % (str(other), self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __div__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the quotient (self / other) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the quotient (self / other) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        (For Python2.x)
        '''
        return self.__truediv__(other)


    def __rdiv__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the quotient (other / self) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the quotient (other / self) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        (For Python2.x)
        '''
        return self.__rtruediv__(other)


    def __pow__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the exponentiation (self ^ other) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the exponentiation (self ^ other) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) ^ (%s)' % (self._definition, other._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '(%s) ^ %s' % (self._definition, str(other))
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __rpow__(self, other):
        '''
        If other is a FerVar, returns an anonymous FerVar whose definition 
        is the exponentiation (other ^ self) of the FerVar definitions.
        If other is Real, returns an anonymous FerVar whose definition 
        is the exponentiation (other ^ self) of the FerVar definition with the number.
        If other is not a FerVar or Real, returns NotImplemented
        '''
        if isinstance(other, FerVar):
            newdef = '(%s) ^ (%s)' % (other._definition, self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            newvar._requires.update(other._requires)
            return newvar
        if isinstance(other, numbers.Real):
            newdef = '%s ^ (%s)' % (str(other), self._definition)
            newvar = FerVar(defn=newdef)
            newvar._requires.update(self._requires)
            return newvar
        return NotImplemented


    def __neg__(self):
        '''
        Returns an anonymous FerVar whose definition is 
        the product of -1.0 and this FerVar definition.
        '''
        newdef = '-1.0 * (%s)' % self._definition
        newvar = FerVar(defn=newdef)
        newvar._requires.update(self._requires)
        return newvar


    def __pos__(self):
        '''
        Returns an anonymous FerVar whose definition is 
        the same as this FerVar definition.
        '''
        newvar = FerVar(defn=self._definition)
        newvar._requires.update(self._requires)
        return newvar


    def __abs__(self):
        '''
        Returns an anonymous FerVar whose definition is 
        the absolute value of this FerVar definition.
        '''
        newdef = 'abs(%s)' % self._definition
        newvar = FerVar(defn=newdef)
        newvar._requires.update(self._requires)
        return newvar


    def __getitem__(self, key):
        '''
        This FerVar must be assigned in Ferret.

        If key is 'data', returns the data array for this FerVar,
        loading it if necessary.
        If key is 'grid', returns the data grid for this FerVar,
        loading it if necessary.
        If key is 'missval', returns the value for missing data 
        for this FerVar.
        If key is 'unit', returns the data unit for this FerVar.

        Otherwise, assumes key is a slice or subset specification, 
        and returns an anonymous FerVar whose definition is a 
        subset of this FerVar.  
            key is an int, float, string, int slice, float slice, 
                string slice, or a tuple of these values.
                 - int are interpreted as index/indices
                 - floats are interpreted as axis values
                 - strings are interpreted as axis values possibly with units
        Units in a string designate an axis; otherwise the index
        within the given tuple (or zero if not a tuple) specifies the axis.
        For example ['20N':'50N'] will always be a latitude subset.

        TODO: handle step values
        '''
        if key is None:
            raise KeyError('None is not a valid key')
        if not self._varname:
            raise NotImplementedError('variable not assigned in Ferret')
        if key == 'data':
           return self.getdata()
        if key == 'grid':
           return self.getgrid()
        if key == 'missval':
           return self.getmissval()
        if key == 'unit':
           return self.getunit()

        coordlimits = [ None ] * pyferret.MAX_FERRET_NDIM
        indexlimits = [ None ] * pyferret.MAX_FERRET_NDIM
        changed = False
        # TODO: handle step values, try to condense code
        if isinstance(key, tuple):
            for k in range(len(key)):
                piece = key[k]
                if piece is None:
                    continue
                if isinstance(piece, slice):
                    try:
                        (axtype, start, stop, step) = pyferret.FerAxis._parsegeoslice(piece)
                    except Exception as ex:
                        raise KeyError('%s is not valid: %s' % (str(piece), str(ex)))
                    if step is not None:
                        raise KeyError('step values in slices are not supported at this time')
                    if axtype == pyferret.AXISTYPE_LONGITUDE:
                        if coordlimits[pyferret.X_AXIS] or indexlimits[pyferret.X_AXIS]:
                            raise KeyError('two longitude slices given')
                        coordlimits[pyferret.X_AXIS] = '%s:%s' % (str(start), str(stop))
                        changed = True
                    elif axtype == pyferret.AXISTYPE_LATITUDE:
                        if coordlimits[pyferret.Y_AXIS] or indexlimits[pyferret.Y_AXIS]:
                            raise KeyError('two latitude slices given')
                        coordlimits[pyferret.Y_AXIS] = '%s:%s' % (str(start), str(stop))
                        changed = True
                    elif axtype == pyferret.AXISTYPE_LEVEL:
                        if coordlimits[pyferret.Z_AXIS] or indexlimits[pyferret.Z_AXIS]:
                            raise KeyError('two level slices given')
                        coordlimits[pyferret.Z_AXIS] = '%s:%s' % (str(start), str(stop))
                        changed = True
                    elif axtype == pyferret.AXISTYPE_TIME:
                        if coordlimits[pyferret.T_AXIS] or indexlimits[pyferret.T_AXIS]:
                            raise KeyError('two time slices given')
                        starttime = pyferret.FerAxis._makedatestring(start)
                        stoptime = pyferret.FerAxis._makedatestring(stop)
                        coordlimits[pyferret.T_AXIS] = '%s:%s' % (starttime, stoptime)
                        changed = True
                    elif isinstance(start,int) and isinstance(stop,int):
                        if coordlimits[k] or indexlimits[k]:
                            raise KeyError('two slices for axis index %d given' % k)
                        # do not know the axis length at this time
                        if (start < 0) or (stop < 0):
                            raise KeyError('negative indices not supported at this time')
                        # Ferret indices start at 1
                        start += 1
                        stop += 1
                        indexlimits[k] = '%d:%d' % (start, stop)
                        changed = True
                    elif isinstance(start,numbers.Real) and isinstance(stop,numbers.Real):
                        if coordlimits[k] or indexlimits[k]:
                            raise KeyError('two slices for axis index %d given' % k)
                        coordlimits[k] = '%s:%s' % (str(start), str(stop))
                        changed = True
                    elif (start is None) and (stop is None):
                        # full range on this axis 
                        if coordlimits[k] or indexlimits[k]:
                            raise KeyError('two slices for axis index %d given' % k)
                        continue
                    else:
                        raise KeyError('%s in not valid' % str(piece))
                else:
                    try:
                        (axtype, val) = pyferret.FerAxis._parsegeoval(piece)
                    except Exception as ex:
                        raise KeyError('%s is not valid: %s' % (str(piece), str(ex)))
                    if axtype == pyferret.AXISTYPE_LONGITUDE:
                        if coordlimits[pyferret.X_AXIS] or indexlimits[pyferret.X_AXIS]:
                            raise KeyError('two longitude slices given')
                        coordlimits[pyferret.X_AXIS] = '%s' % str(val)
                        changed = True
                    elif axtype == pyferret.AXISTYPE_LATITUDE:
                        if coordlimits[pyferret.Y_AXIS] or indexlimits[pyferret.Y_AXIS]:
                            raise KeyError('two latitude slices given')
                        coordlimits[pyferret.Y_AXIS] = '%s' % str(val)
                        changed = True
                    elif axtype == pyferret.AXISTYPE_LEVEL:
                        if coordlimits[pyferret.Z_AXIS] or indexlimits[pyferret.Z_AXIS]:
                            raise KeyError('two level slices given')
                        coordlimits[pyferret.Z_AXIS] = '%s' % str(val)
                        changed = True
                    elif axtype == pyferret.AXISTYPE_TIME:
                        if coordlimits[pyferret.T_AXIS] or indexlimits[pyferret.T_AXIS]:
                            raise KeyError('two time slices given')
                        coordlimits[pyferret.T_AXIS] = pyferret.FerAxis._makedatestring(val)
                        changed = True
                    elif isinstance(val,int):
                        if coordlimits[k] or indexlimits[k]:
                            raise KeyError('two slices for axis index %d given' % k)
                        # do not know the axis length at this time
                        if val < 0: 
                            raise KeyError('negative indices not supported at this time')
                        # Ferret indices start at 1
                        val += 1
                        indexlimits[k] = '%d' % val
                        changed = True
                    elif isinstance(val,numbers.Real):
                        if coordlimits[k] or indexlimits[k]:
                            raise KeyError('two slices for axis index %d given' % k)
                        coordlimits[k] = '%s' % str(val)
                        changed = True
                    else:
                        raise KeyError('%s in not valid' % str(piece))
        elif isinstance(key, slice):
            try:
                (axtype, start, stop, step) = pyferret.FerAxis._parsegeoslice(key)
            except Exception as ex:
                raise KeyError('%s is not valid: %s' % (str(key), str(ex)))
            if step is not None:
                raise KeyError('step values in slices are not supported at this time')
            if axtype == pyferret.AXISTYPE_LONGITUDE:
                coordlimits[pyferret.X_AXIS] = '%s:%s' % (str(start), str(stop))
                changed = True
            elif axtype == pyferret.AXISTYPE_LATITUDE:
                coordlimits[pyferret.Y_AXIS] = '%s:%s' % (str(start), str(stop))
                changed = True
            elif axtype == pyferret.AXISTYPE_LEVEL:
                coordlimits[pyferret.Z_AXIS] = '%s:%s' % (str(start), str(stop))
                changed = True
            elif axtype == pyferret.AXISTYPE_TIME:
                starttime = pyferret.FerAxis._makedatestring(start)
                stoptime = pyferret.FerAxis._makedatestring(stop)
                coordlimits[pyferret.T_AXIS] = '%s:%s' % (starttime, stoptime)
                changed = True
            elif isinstance(start,int) and isinstance(stop,int):
                # do not know the axis length at this time
                if (start < 0) or (stop < 0):
                    raise KeyError('negative indices not supported at this time')
                # Ferret indices start at 1
                start += 1
                stop += 1
                indexlimits[0] = '%d:%d' % (start, stop)
                changed = True
            elif isinstance(start,numbers.Real) and isinstance(stop,numbers.Real):
                coordlimits[0] = '%s:%s' % (str(start), str(stop))
                changed = True
            elif (start is None) and (stop is None):
                # full range - standard way of generating a duplicate
                pass
            else:
                raise KeyError('%s in not valid' % str(key))
        else:
            try:
                (axtype, val) = pyferret.FerAxis._parsegeoval(key)
            except Exception as ex:
                raise KeyError('%s is not valid: %s' % (str(key), str(ex)))
            if axtype == pyferret.AXISTYPE_LONGITUDE:
                coordlimits[pyferret.X_AXIS] = '%s' % str(val)
                changed = True
            elif axtype == pyferret.AXISTYPE_LATITUDE:
                coordlimits[pyferret.Y_AXIS] = '%s' % str(val)
                changed = True
            elif axtype == pyferret.AXISTYPE_LEVEL:
                coordlimits[pyferret.Z_AXIS] = '%s' % str(val)
                changed = True
            elif axtype == pyferret.AXISTYPE_TIME:
                coordlimits[pyferret.T_AXIS] = pyferret.FerAxis._makedatestring(val)
                changed = True
            elif isinstance(val,int):
                # do not know the axis length at this time
                if val < 0: 
                    raise KeyError('negative indices not supported at this time')
                # Ferret indices start at 1
                val += 1
                indexlimits[k] = '%d' % val
                changed = True
            elif isinstance(start,numbers.Real):
                coordlimits[k] = '%s' % str(val)
                changed = True
            else:
                raise KeyError('%s in not valid' % str(key))
        if not changed:
            # the whole thing - definition is just this variable
            newvar = FerVar(defn=self.fername())
            newvar._requires.update(self._requires)
            return newvar
        # create the subset definition in Ferret
        if self._dsetname:
            newdef = '%s[d=%s,' % (self._varname, self._dsetname)
        else:
            newdef = '%s[' % self._varname
        if coordlimits[pyferret.X_AXIS]:
            newdef += 'X=%s,' % coordlimits[pyferret.X_AXIS]
        if indexlimits[pyferret.X_AXIS]:
            newdef += 'I=%s,' % indexlimits[pyferret.X_AXIS]
        if coordlimits[pyferret.Y_AXIS]:
            newdef += 'Y=%s,' % coordlimits[pyferret.Y_AXIS]
        if indexlimits[pyferret.Y_AXIS]:
            newdef += 'J=%s,' % indexlimits[pyferret.Y_AXIS]
        if coordlimits[pyferret.Z_AXIS]:
            newdef += 'Z=%s,' % coordlimits[pyferret.Z_AXIS]
        if indexlimits[pyferret.Z_AXIS]:
            newdef += 'K=%s,' % indexlimits[pyferret.Z_AXIS]
        if coordlimits[pyferret.T_AXIS]:
            newdef += 'T=%s,' % coordlimits[pyferret.T_AXIS]
        if indexlimits[pyferret.T_AXIS]:
            newdef += 'L=%s,' % indexlimits[pyferret.T_AXIS]
        if coordlimits[pyferret.E_AXIS]:
            newdef += 'E=%s,' % coordlimits[pyferret.E_AXIS]
        if indexlimits[pyferret.E_AXIS]:
            newdef += 'M=%s,' % indexlimits[pyferret.E_AXIS]
        if coordlimits[pyferret.F_AXIS]:
            newdef += 'F=%s,' % coordlimits[pyferret.F_AXIS]
        if indexlimits[pyferret.F_AXIS]:
            newdef += 'N=%s,' % indexlimits[pyferret.F_AXIS]
        # replace the final , with ]
        newdef = newdef[:-1] + ']'
        newvar = FerVar(defn=newdef)
        newvar._requires.update(self._requires)
        return newvar


    def __getattr__(self, name):
        '''
        Return the data array (if name='data'), data grid (if name='grid'), 
        name (if name='name'), or a copy of the coordinates (if name='coords')
        Note that this method is only called when the parent object 
        does not have an attribute with this name.
        '''
        try:
            if name in _ADDED_ATTRIBUTES:
                return self.__getitem__(name)
        except KeyError:
            pass
        raise AttributeError("unknown attribute '%s'" % name)


    def __dir__(self):
        '''
        Returns a list of known attributes, including those added 
        by the __getattr__ method.
        '''
        mydir = list(_ADDED_ATTRIBUTES)
        mydir.extend( dir(super(FerVar, self)) )
        return mydir


    def _markasknownvar(self, varname, dsetname, isfilevar):
        '''
        Marks this variable as a variable already defined in Ferret.
        '''
        if not varname:
            raise ValueError('varname is not given')
        if not isinstance(varname, str):
            raise ValueError('varname is not a string')
        if dsetname and not isinstance(varname, str):
            raise ValueError('dsetname name is not a string')
        self._varname = varname
        if dsetname:
            self._dsetname = dsetname
        else:
            self._dsetname = ''
        self._isfilevar = bool(isfilevar)
        self._definition = self.fername()
        self._requires.add(varname.upper())
        self.unload()


    def _assigninferret(self, varname, dsetname):
        '''
        Defines this FerVar in Ferret using the given variable name 
        associated with the given dataset name.
            varname (string): name for the variable in Ferret
            dsetname (string): name of the dataset to contain the variable
        Raises a ValueError if there is a problem.
        '''
        if not self._definition:
            raise ValueError('this FerVar does not contain a definition')
        if not varname:
            raise ValueError('variable name to be assigned is not given')
        if varname.upper() in self._requires:
            raise ValueError('recursive definitions cannot be implemented in Ferret')
        # Assign the variable in Ferret
        cmdstr = 'DEFINE VAR'
        if dsetname:
            cmdstr += '/D="%s"' % dsetname
        if self._title:
            cmdstr += '/TITLE="%s"' % self._title
        cmdstr += ' %s = %s' % (varname, self._definition)
        (errval, errmsg) = pyferret.run(cmdstr)
        if errval != pyferret.FERR_OK:
            raise ValueError('problems defining %s (%s) in Ferret: %s' % (varname, cmdstr, errmsg))
        # Revise the fields in this FerVar to reflect this assignment
        self._markasknownvar(varname, dsetname, False)


    def _removefromferret(self):
        '''
        Removes (cancels) this variable in Ferret, then unloads this FerVar 
        and erases _varname.  Raises a NotImplementedError is this is a file 
        variable.  Raises a ValueError if there is a Ferret problem.  This 
        normally is not called by the user; instead delete the FerVar from 
        the dataset.
        '''
        # ignore if this Ferrer variable has already been removed from Ferret
        if not self._varname:
            return
        fername = self.fername()
        if self._isfilevar:
            raise NotImplementedError('%s is a file variable; close the dataset to remove' % fername)
        cmdstr = 'CANCEL VAR %s' % fername
        (errval, errmsg) = pyferret.run(cmdstr)
        if errval != pyferret.FERR_OK:
            raise ValueError('unable to remove variable %s from Ferret: %s' % (fername, errmsg))
        self._varname = ''
        self.unload()


    def unload(self):
        '''
        Clears the grid and data stored in this FerVar.  After this call, any 
        request for the grid or data will automatically load the latest values 
        from Ferret.  This method should be called anytime there is a change 
        in the definition of this variable, or a variable this variable uses.
        '''
        self._datagrid = None
        self._dataarray = None
        self._dataunit = ''
        self._missingvalue = None


    def load(self):
        '''
        Retrieves the grid and data for this Ferret variable from Ferret.
        This method is automatically called before returning the grid or data 
        for the first time for this variable.  This can be called to update
        the grid or data in this FerVar after any change in the definition 
        of the variable.  Alternatively, cleardata can be called to clear any
        stored grid and data, delaying the update from Ferret until the grid
        or data is requested.
        Raises a ValueEror if problems occur.
        '''
        fername = self.fername()
        datadict = pyferret.getdata(fername, False)
        feraxes = [ ]
        for (axistype,axcoords,axunit,axname) in zip(
                datadict["axis_types"], datadict["axis_coords"], 
                datadict["axis_units"], datadict["axis_names"]):
            feraxes.append( pyferret.FerAxis(coords=axcoords, 
                    axtype=axistype, unit=axunit, name=axname) )
        self._datagrid = pyferret.FerGrid(axes=feraxes, name=fername)
        self._dataarray = datadict["data"]
        self._dataunit = datadict["data_unit"]
        self._missingvalue = datadict["missing_value"]


    def getdata(self):
        '''
        Returns a copy of the data array for this Ferret variable,
        first loading this variable if necessary.
        Raises a ValueError is a problem occurs.
        '''
        if (self._datagrid is None) or (self._dataarray is None):
            self.load()
        return self._dataarray.copy('A')


    def getgrid(self):
        '''
        Returns a copy of the data grid for this Ferret variable,
        first loading this variable if necessary.
        Raises a ValueError is a problem occurs.
        '''
        if (self._datagrid is None) or (self._dataarray is None):
            self.load()
        return self._datagrid.copy()


    def getmissval(self):
        '''
        Returns the value used for missing data for this Ferret 
        variable, first loading this variable if necessary.  
        Raises a ValueError is a problem occurs.
        '''
        if (self._datagrid is None) or (self._dataarray is None):
            self.load()
        # The missing value is a single-element ndarray
        return self._missingvalue[0]


    def getunit(self):
        '''
        Returns the unit string of the data for this Ferret
        variable, first loading this variable if necessary.
        Raises a ValueError is a problem occurs.
        '''
        if (self._datagrid is None) or (self._dataarray is None):
            self.load()
        return self._dataunit


    def showgrid(self, qual=''):
        '''
        Show the Ferret grid information about this variable.  This uses 
        the Ferret SHOW GRID command to create and display the information.
            qual (string): Ferret qualifiers to add to the SHOW GRID command
        '''
        if not isinstance(qual, str):
            raise ValueError('qual (Ferret qualifiers) must be a string')
        cmdstr = 'SHOW GRID'
        if qual:
            cmdstr += qual
        cmdstr += ' '
        cmdstr += self.fername()
        (errval, errmsg) = pyferret.run(cmdstr)
        if errval != pyferret.FERR_OK:
            raise ValueError('Ferret command "%s" failed: %s' % (cmdstr, errmsg))


    def regrid(self, newgrid, method=REGRID_LINEAR):
        '''
        Returns an anonymous FerVar that is this variable regridded to the grid
        implied by newgrid using the given method.
            newgrid (FerVar |  string | FerGrid): regrid to this implied grid;
                if a FerVar, the implied grid is the grid used by the Ferret variable,
                if a string, the implied grid is the grid known to Ferret by this name
                if a FerGrid, the implied grid is this grid (TODO: implement)
            method (string): method to perform the regridding; typically one of
                pyferret.REGRID_LINEAR (default)
                    (multi-axis) linear interpolation of nearest source points around destination point
                pyferret.REGRID_AVERAGE
                    length-weighted averaging of source point cells overlapping destination point cell
                pyferret.REGRID_ASSOCIATE
                    blind association of source points to destination points by indices
                pyferret.REGRID_MEAN
                    unweighted mean of source points in destination point cell
                pyferret.REGRID_NEAREST
                    value of source point nearest the destination point
                pyferret.REGRID_MIN
                    minimum value of source points in destination point cell 
                pyferret.REGRID_MAX
                    maximum value of source points in destination point cell 
                pyferret.REGRID_EXACT
                    copy values where source and destination points coincide; 
                    other destination points assigned missing value
        '''
        if not self._varname:
            raise NotImplementedError('regridding can only be performed on variables assigned in Ferret')
        if not ( isinstance(method, str) and (method[0] == '@') ):
            raise ValueError('invalid regridding method %s' % str(method))
        if isinstance(newgrid, FerVar):
            if not newgrid._varname:
                raise ValueError('FerVar used for the new grid is not assigned in Ferret')
            gridname = newgrid.fername()
        elif isinstance(newgrid, str):
            gridname = newgrid
        elif isinstance(newgrid, pyferret.FerGrid):
            raise NotImplementedError('regrid using FerGrid not implemented at this time')
        if self._dsetname:
            newdef = '%s[d=%s,g=%s%s]' % (self._varname, self._dsetname, gridname, method)
        else:
            newdef = '%s[g=%s%s]' % (self._varname, gridname, method)
        newvar = FerVar(defn=newdef)
        newvar._requires.update(self._requires)
        return newvar




'''
Tests of the regrid.__init__ functions

@author: Karl Smith
'''
import unittest
import numpy
from . import __init__ as regrid

class RegridTests(unittest.TestCase):
    '''
    Tests of the regrid.__init__ functions
    '''


    def testQuadCornersFrom3D(self):
        '''
        Test of regrid.quadCornersFrom3D
        '''
        xedges = numpy.linspace(2.0, 10.0, 5)
        yedges = numpy.linspace(15.0, 75.0, 7)
        expect_ptsy, expect_ptsx = numpy.meshgrid(yedges, xedges)

        ptsx3d = numpy.empty((xedges.shape[0] - 1, yedges.shape[0] - 1, 4))
        ptsy3d = numpy.empty((xedges.shape[0] - 1, yedges.shape[0] - 1, 4))

        ptsx3d[:, :, 0] = expect_ptsx[:-1, :-1]
        ptsx3d[:, :, 1] = expect_ptsx[1:, :-1]
        ptsx3d[:, :, 2] = expect_ptsx[1:, 1:]
        ptsx3d[:, :, 3] = expect_ptsx[:-1, 1:]
        ptsy3d[:, :, 0] = expect_ptsy[:-1, :-1]
        ptsy3d[:, :, 1] = expect_ptsy[1:, :-1]
        ptsy3d[:, :, 2] = expect_ptsy[1:, 1:]
        ptsy3d[:, :, 3] = expect_ptsy[:-1, 1:]
        ptsx, ptsy = regrid.quadCornersFrom3D(ptsx3d, ptsy3d)
        self.assertTrue(numpy.allclose(expect_ptsx, ptsx),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ptsx), str(ptsx)))
        self.assertTrue(numpy.allclose(expect_ptsy, ptsy),
                        "Expected Y coordinates:\n%s\nFound Y coordinates:\n%s" % \
                        (str(expect_ptsy), str(ptsy)))

        ptsx3d[:, :, 0] = expect_ptsx[:-1, :-1]
        ptsx3d[:, :, 3] = expect_ptsx[1:, :-1]
        ptsx3d[:, :, 2] = expect_ptsx[1:, 1:]
        ptsx3d[:, :, 1] = expect_ptsx[:-1, 1:]
        ptsy3d[:, :, 0] = expect_ptsy[:-1, :-1]
        ptsy3d[:, :, 3] = expect_ptsy[1:, :-1]
        ptsy3d[:, :, 2] = expect_ptsy[1:, 1:]
        ptsy3d[:, :, 1] = expect_ptsy[:-1, 1:]
        ptsx, ptsy = regrid.quadCornersFrom3D(ptsx3d, ptsy3d)
        self.assertTrue(numpy.allclose(expect_ptsx, ptsx),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ptsx), str(ptsx)))
        self.assertTrue(numpy.allclose(expect_ptsy, ptsy),
                        "Expected Y coordinates:\n%s\nFound Y coordinates:\n%s" % \
                        (str(expect_ptsy), str(ptsy)))


    def testQuadCentroids(self):
        '''
        Test of regrid.quadCentroids
        '''
        xedges = numpy.linspace(2.0, 10.0, 5)
        xcenters = 0.5 * (xedges[:-1] + xedges[1:])
        yedges = numpy.linspace(15.0, 75.0, 7)
        ycenters = 0.5 * (yedges[:-1] + yedges[1:])
        ptsx, ptsy = numpy.meshgrid(xedges, yedges)
        expect_ctrx, expect_ctry = numpy.meshgrid(xcenters, ycenters)
        ctrx, ctry = regrid.quadCentroids(ptsx, ptsy)
        self.assertTrue(numpy.allclose(expect_ctrx, ctrx),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ctrx), str(ctrx)))
        self.assertTrue(numpy.allclose(expect_ctry, ctry),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ctry), str(ctry)))

        for k in range(1, ptsx.shape[0]):
            ptsx[k, :] += 3.0 * k
        for k in range(expect_ctrx.shape[0]):
            expect_ctrx[k, :] += 1.5 * (2 * k + 1)
        for k in range(1, ptsy.shape[1]):
            ptsy[:, k] += 4.0 * k
        for k in range(expect_ctry.shape[1]):
            expect_ctry[:, k] += 2.0 * (2 * k + 1)
        ctrx, ctry = regrid.quadCentroids(ptsx, ptsy)
        self.assertTrue(numpy.allclose(expect_ctrx, ctrx),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ctrx), str(ctrx)))
        self.assertTrue(numpy.allclose(expect_ctry, ctry),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ctry), str(ctry)))

        for k in range(1, ptsx.shape[1]):
            ptsx[:, k] += 5.0 * k
        for k in range(expect_ctrx.shape[1]):
            expect_ctrx[:, k] += 2.5 * (2 * k + 1)
        for k in range(1, ptsy.shape[0]):
            ptsy[k, :] += 7.0 * k
        for k in range(expect_ctry.shape[0]):
            expect_ctry[k, :] += 3.5 * (2 * k + 1)
        ctrx, ctry = regrid.quadCentroids(ptsx, ptsy)
        self.assertTrue(numpy.allclose(expect_ctrx, ctrx),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ctrx), str(ctrx)))
        self.assertTrue(numpy.allclose(expect_ctry, ctry),
                        "Expected X coordinates:\n%s\nFound X coordinates:\n%s" % \
                        (str(expect_ctry), str(ctry)))



if __name__ == "__main__":
    unittest.main()



'''
Representation of a space-time region of interest
'''

import numbers
import pyferret

class FerRegion(object):
    '''
    A logically rectagular region specified using axis limit qualifiers.
    Also used as a superclass to provide a bounding box for more elaborate 
    regions.
    '''

    def __init__(self, X=None, Y=None, Z=None, T=None, E=None, F=None,
                       I=None, J=None, K=None, L=None, M=None, N=None, qual=''):
        '''
        A logically rectagular regions specified using axis limit qualifiers;
        eg; X="90W:20E",Y="20N:60N"

        The X, Y, Z, T, E, and F arguments refer to an axis coordinate value 
        or range of axis coordinate values.  Coordinate values are interpreted 
        according to Ferret syntax rules (ranges can include the endpoint).
            X (float, str, or slice of float or str): 
                X (longitude) axis coordinate position or range
            Y (float, str, or slice of float or str): 
                Y (latitude) axis coordinate position or range
            Z (float, str, or slice of float or str): 
                Z (level) axis coordinate position or range
            T (float, str, or slice of float or str): 
                T (time) axis coordinate position or range
            E (float, str, or slice of float or str): 
                E (ensemble) axis coordinate position or range
            F (float, str, or slice of float or str): 
                F (forecast) axis coordinate position or range

        The I, J, K, L, M, and N arguments refer to an axis coordinate index 
        or range of indices.  Integer values are interpreted as python indices 
        (starts with zero, excludes endpoint in ranges).  String values are 
        used as-is and thus are interpreted according to Ferret syntax rules 
        (starts at one, includes endpoint in ranges).
            I (int, str, or slice of int or str): 
                X (longitude) axis index or range of indices
            J (int, str, or slice of int or str): 
                Y (latitude) axis index or range of indices
            K (int, str, or slice of int or str): 
                Z (level) axis index or range of indices
            L (int, str, or slice of int or str): 
                T (time) axis index or range of indices
            M (int, str, or slice of int or str): 
                E (ensemble) axis index or range of indices
            N (int, str, or slice of int or str): 
                F (forecast) axis index or range of indices

        For any axis, either a coordinate or an index specification can be
        given, but not both.

        The qual argument is a string giving any additional Ferret-syntax qualifiers.
        '''
        # coordinate qualifiers
        self._xaxisqual = self._interpcoord('X', X)
        self._yaxisqual = self._interpcoord('Y', Y)
        self._zaxisqual = self._interpcoord('Z', Z)
        self._taxisqual = self._interpcoord('T', T)
        self._eaxisqual = self._interpcoord('E', E)
        self._faxisqual = self._interpcoord('F', F)
        # index qualifiers
        idxqual = self._interpindex('I', I)
        if idxqual:
            if self._xaxisqual:
                raise ValueError('X and I cannot both be given')
            self._xaxisqual = idxqual
        idxqual = self._interpindex('J', J)
        if idxqual:
            if self._yaxisqual:
                raise ValueError('Y and J cannot both be given')
            self._yaxisqual = idxqual
        idxqual = self._interpindex('K', K)
        if idxqual:
            if self._zaxisqual:
                raise ValueError('Z and K cannot both be given')
            self._zaxisqual = idxqual
        idxqual = self._interpindex('L', L)
        if idxqual:
            if self._taxisqual:
                raise ValueError('T and L cannot both be given')
            self._taxisqual = idxqual
        idxqual = self._interpindex('M', M)
        if idxqual:
            if self._eaxisqual:
                raise ValueError('E and M cannot both be given')
            self._eaxisqual = idxqual
        idxqual = self._interpindex('N', N)
        if idxqual:
            if self._faxisqual:
                raise ValueError('F and N cannot both be given')
            self._faxisqual = idxqual
        # additional qualifiers
        if not isinstance(qual, str):
            raise ValueError('qual must be a string')
        self._addnlqual = qual.strip()
        # masking variable for more complex regions (for superclasses)
        self._maskvar = None


    def __repr__(self):
        '''
        Representation of this FerRegion.  Arguments are always given in the
        string representation (thus, Ferret syntax for indices and ranges).
        '''
        repstr = 'FerRegion('
        if self._xaxisqual.startswith('/X='):
            repstr += "X='" + self._xaxisqual[3:] + "',"
        elif self._xaxisqual.startswith('/I='):
            repstr += "I='" + self._xaxisqual[3:] + "',"
        if self._yaxisqual.startswith('/Y='):
            repstr += "Y='" + self._yaxisqual[3:] + "',"
        elif self._yaxisqual.startswith('/J='):
            repstr += "J='" + self._yaxisqual[3:] + "',"
        if self._zaxisqual.startswith('/Z='):
            repstr += "Z='" + self._zaxisqual[3:] + "',"
        elif self._zaxisqual.startswith('/K='):
            repstr += "K='" + self._zaxisqual[3:] + "',"
        if self._taxisqual.startswith('/T='):
            repstr += "T='" + self._taxisqual[3:] + "',"
        elif self._taxisqual.startswith('/L='):
            repstr += "L='" + self._taxisqual[3:] + "',"
        if self._eaxisqual.startswith('/E='):
            repstr += "E='" + self._eaxisqual[3:] + "',"
        elif self._eaxisqual.startswith('/M='):
            repstr += "M='" + self._eaxisqual[3:] + "',"
        if self._faxisqual.startswith('/F='):
            repstr += "F='" + self._faxisqual[3:] + "',"
        elif self._faxisqual.startswith('/N='):
            repstr += "N='" + self._faxisqual[3:] + "',"
        if self._addnlqual:
            repstr += "qual='" + self._addnlqual + "'"
        if repstr.endswith(","):
            repstr = repstr[:-1]
        repstr += ")"
        return repstr


    def _ferretqualifierstr(self):
        '''
        Returns the Ferret syntax axis qualifier string describing this region;
        eg, '/X=90W:20E/Y=20N:60N'
        If no axis qualifiers are given, an empty string is returned.
        '''
        qualstr  = self._xaxisqual
        qualstr += self._yaxisqual
        qualstr += self._zaxisqual
        qualstr += self._taxisqual
        qualstr += self._eaxisqual
        qualstr += self._faxisqual
        qualstr += self._addnlqual
        return qualstr


    @staticmethod
    def _interpcoord(name, value):
        '''
        Converts a coordinate value or range specification to a Ferret coordinate qualifier
            name ('X','Y','Z','T','E','F'): name of the value for the qualifier
            value (number, str, number slice, or str slice): value to interpret
        Returns (str): Ferret coordinate qualifier ('/X=165')
        If value is None, name is ignored and an empty string is returned.
        Raises ValueError if unable to interpret value.
        '''
        if value is None:
            return ''
        if not name in ('X','Y','Z','T','E','F'):
            raise ValueError('Invalid variable name "' + str(name) + '"')
        qual = '/' + name + '='
        if isinstance(value, numbers.Real):
            return qual + str(value)
        elif isinstance(value, str):
            val = value.strip()
            if val:
                return qual + val
        elif isinstance(value, slice):
            start = value.start
            stop = value.stop
            step = value.step
            if isinstance(start, numbers.Real) and isinstance(stop, numbers.Real) and (step is None):
                return qual + str(start) + ':' + str(stop)
            elif isinstance(start, numbers.Real) and isinstance(stop, numbers.Real) and isinstance(step, numbers.Real):
                return qual + str(start) + ':' + str(stop) + ':' + str(step)
            elif isinstance(start, str) and isinstance(stop, str) and (step is None):
                start = start.strip()
                stop = stop.strip()
                if start and stop:
                    return qual + start + ':' + stop
            elif isinstance(start, str) and isinstance(stop, str) and isinstance(step, str):
                start = start.strip()
                stop = stop.strip()
                step = step.strip()
                if start and stop and step:
                    return qual + start + ':' + stop + ':' + step
        raise ValueError('definition for ' + name + ' is invalid')


    @staticmethod
    def _interpindex(name, value):
        '''
        Converts a coordinate index or range specification to a Ferret coordinate qualifier
        Int values and slices are interpreted as Python syntax (start at zero, ranges exclude
        the endpoint).  String values and slices are interpreted as Ferret syntax (start at
        one, ranges include endpoint).
            name ('I','J','K','L','M','N'): name of the value for the qualifier
            value (int, str, int slice, or str slice): value to interpret
        Returns (str): Ferret coordinate qualifier ('/I=1:10')
        Raises ValueError if unable to interpret value.
        '''
        if value is None:
            return ''
        if not name in ('I','J','K','L','M','N'):
            raise ValueError('Invalid variable name "' + str(name) + '"')
        qual = '/' + name + '='
        if isinstance(value, numbers.Integral):
            return qual + str(value + 1)
        elif isinstance(value, str):
            val = value.strip()
            if val:
                return qual + val
        elif isinstance(value, slice):
            start = value.start
            stop = value.stop
            step = value.step
            if isinstance(start, numbers.Integral) and isinstance(stop, numbers.Integral) and (step is None):
                return qual + str(start + 1) + ':' + str(stop)
            elif isinstance(start, numbers.Integral) and isinstance(stop, numbers.Integral) and isinstance(step, numbers.Integral):
                return qual + str(start + 1) + ':' + str(stop) + ':' + str(step)
            elif isinstance(start, str) and isinstance(stop, str) and (step is None):
                start = start.strip()
                stop = stop.strip()
                if start and stop:
                    return qual + start + ':' + stop
            elif isinstance(start, str) and isinstance(stop, str) and isinstance(step, str):
                start = start.strip()
                stop = stop.strip()
                step = step.strip()
                if start and stop and step:
                    return qual + start + ':' + stop + ':' + step
        raise ValueError('definition for ' + name + ' is invalid')





"""
Performs a chi-square test that a sample with the observed
counts of categorical data comes from a population with the
given expected counts or relative frequencies of that data.
"""

from __future__ import print_function

import numpy
import pyferret
import scipy.stats


def ferret_init(id):
    """
    Initialization for the stats_chisquare Ferret PyEF
    """
    axes_values = [ pyferret.AXIS_DOES_NOT_EXIST ] * pyferret.MAX_FERRET_NDIM
    axes_values[0] = pyferret.AXIS_CUSTOM
    false_influences = [ False ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 3,
                "descript": "Returns chi-square test stat. and prob. (and num. good categories, N) " \
                            "that sample counts of cat. data matches pop. expected counts. ",
                "axes": axes_values,
                "argnames": ( "SAMPLE_CNTS", "EXPECT_CNTS", "DELTA_DEGFREE", ),
                "argdescripts": ( "Sample counts of categorical data",
                                  "Expected counts or relative frequencies (will be adjusted)",
                                  "Difference from standard (N-1) degrees of freedom (num. computed parameters)", ),
                "argtypes": ( pyferret.FLOAT_ARRAY, pyferret.FLOAT_ARRAY, pyferret.FLOAT_ONEVAL, ),
                "influences": ( false_influences, false_influences, false_influences, ),
              }
    return retdict


def ferret_custom_axes(id):
    """
    Define custom axis of the stats_chisquare Ferret PyEF
    """
    axis_defs = [ None ] * pyferret.MAX_FERRET_NDIM
    axis_defs[0] = ( 1, 3, 1, "X2,P,N", False )
    return axis_defs


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Performs a chi-square test that a sample with the observed counts
    of categorical data, given in inputs[0], comes from a population
    with the expected counts or relative frequencies given in inputs[1].
    The difference from the standard (n-1) degrees of freedom (eg, the
    number of population parameters estimated from the sample) are given
    in inputs[2].  The test statistic value and probability are returned
    in result.  The counts arrays must either have the same shape or both
    have a single defined non-sigular axis of the same size.  Categories
    that contain undefined counts are elimated before performing the
    test.  Population values that are used in the test are adjust so
    their sum equals the sum of the sample counts used in the test.
    """
    if inputs[0].shape != inputs[1].shape :
        shp0 = inputs[0].squeeze().shape
        shp1 = inputs[1].squeeze().shape
        if (len(shp0) > 1) or (len(shp1) > 1) or (shp0 != shp1):
            raise ValueError("SAMPLE_CNTS and EXPECT_CNTS must either have identical dimensions " \
                             "or both have only one defined non-singular axis of the same length")
    samcnts = inputs[0].reshape(-1)
    popcnts = inputs[1].reshape(-1)
    badsam = ( numpy.fabs(samcnts - inpbdfs[0]) < 1.0E-5 )
    badsam = numpy.logical_or(badsam, numpy.isnan(samcnts))
    goodsam = numpy.logical_not(badsam)
    badpop = ( numpy.fabs(popcnts - inpbdfs[1]) < 1.0E-5 )
    badpop = numpy.logical_or(badpop, numpy.isnan(popcnts))
    goodpop = numpy.logical_not(badpop)
    goodmask = numpy.logical_and(goodsam, goodpop)
    samcnts = numpy.array(samcnts[goodmask], dtype=numpy.float64)
    numgood = len(samcnts)
    if numgood < 2:
        raise ValueError("Not enough defined counts in common in SAMPLE_CNTS and EXPECT_CNTS")
    popcnts = numpy.array(popcnts[goodmask], dtype=numpy.float64)
    # Adjust the expected counts so its sum matches the sum of the sample
    # counts;  thus expected counts can be proportions instead of counts
    # and removes issues about missing values.  Get the adjustment factor
    # from the means instead of the sums for accuracy.
    popcnts = popcnts * (samcnts.mean() / popcnts.mean())
    ddof = int(float(inputs[2]) + 0.5)
    fitparams = scipy.stats.chisquare(samcnts, popcnts, ddof)
    result[:] = resbdf
    # chi-square test statistic
    result[0] = fitparams[0]
    # probability
    result[1] = fitparams[1]
    # number of good categories
    result[2] = numgood


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init and ferret_custom_axes do not have problems
    info = ferret_init(0)
    info = ferret_custom_axes(0)

    # Get a sample histogram and expected frequencies
    ddof  = 3
    nbins = 90
    ssize = 100 * nbins
    distf = scipy.stats.weibull_min(2.0, 5.0)
    chival = 1000.0
    while chival > 100.0:
        sample = distf.rvs(ssize)
        bedges = distf.isf(numpy.linspace(0.95,0.05,nbins+1))
        (histgr, retedges) = numpy.histogram(sample, bins=bedges)
        histgr = numpy.array(histgr, dtype=numpy.float64)
        exphist = numpy.ones((nbins,), dtype=numpy.float64) * histgr.mean()
        chival = ((histgr - exphist)**2 / exphist).sum()
        print("created a sample with chival = %f" % chival)
    prob = scipy.stats.chi2(nbins - 1 - ddof).sf(chival)
    expect = numpy.array([chival, prob, nbins], dtype=numpy.float64)
    print("sample histogram = \n%s" % str(histgr))
    print("expect histogram value for all bins = %f" % exphist[0])
    print("expect result = %s" % str(expect))

    # setup for the call to ferret_compute - one non-singular axis
    inpbdfs = numpy.array([-9999.0, -8888.0, -7777.0], dtype=numpy.float64)
    resbdf = numpy.array([-6666.0], dtype=numpy.float64)
    samhist = inpbdfs[0] * numpy.ones((1, 1, 2 * nbins, 1, 1, 1), dtype=numpy.float64, order='F')
    samhist[0, 0, ::2, 0, 0, 0] = histgr
    pophist = numpy.ones((1, 2 * nbins, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    ddofarr = numpy.array([ddof], dtype=numpy.float64)
    result = -5555.0 * numpy.ones((3, 1, 1, 1, 1, 1), dtype=numpy.float64, order='F')

    # call ferret_compute and check the result
    ferret_compute(0, result, resbdf, (samhist, pophist, ddofarr), inpbdfs)
    result = result.reshape(-1)
    print(" found result = %s" % str(result))
    if not numpy.allclose(result, expect):
        raise ValueError("Unexpected result")

    # setup for the call to ferret_compute - multiple dimensions
    inpbdfs = numpy.array([-9999.0, -8888.0, -7777.0], dtype=numpy.float64)
    resbdf = numpy.array([-6666.0], dtype=numpy.float64)
    samhist = inpbdfs[0] * numpy.ones((1, 2, nbins, 1, 1, 1), dtype=numpy.float64, order='F')
    samhist[0, 0, ::2, 0, 0, 0] = histgr[0:nbins//2]
    samhist[0, 1, 1::2, 0, 0, 0] = histgr[nbins//2:]
    pophist = numpy.ones((1, 2, nbins, 1, 1, 1), dtype=numpy.float64, order='F')
    ddofarr = numpy.array([ddof], dtype=numpy.float64)
    result = -5555.0 * numpy.ones((3, 1, 1, 1, 1, 1), dtype=numpy.float64, order='F')

    # call ferret_compute and check the result
    ferret_compute(0, result, resbdf, (samhist, pophist, ddofarr), inpbdfs)
    result = result.reshape(-1)
    print(" found result = %s" % str(result))
    if not numpy.allclose(result, expect):
        raise ValueError("Unexpected result")

    # All successful
    print("Success")

